@@ -28,10 +28,9 @@ Approach:
 """
 
 from urllib.error import (URLError, HTTPError)
-from urllib.parse import (quote, urlparse)
+from urllib.parse import urlparse
 from urllib.request import urlopen
 import codecs
-import hashlib
 import json
 import os.path
 import re
@@ -51,6 +50,7 @@ from ubuntutools.lp.lpapicache import (Launchpad, Distribution, PersonTeam,
 from ubuntutools.lp.udtexceptions import (PackageNotFoundException,
                                           SeriesNotFoundException,
                                           InvalidDistroValueError)
+from ubuntutools.misc import (download, verify_file_checksum)
 from ubuntutools.version import Version
 
 import logging
@@ -81,21 +81,11 @@ class Dsc(debian.deb822.Dsc):
 
     def verify_file(self, pathname):
         "Verify that pathname matches the checksums in the dsc"
-        if os.path.isfile(pathname):
-            alg, checksums = self.get_strongest_checksum()
-            size, digest = checksums[os.path.basename(pathname)]
-            if os.path.getsize(pathname) != size:
-                return False
-            hash_func = getattr(hashlib, alg)()
-            f = open(pathname, 'rb')
-            while True:
-                buf = f.read(hash_func.block_size)
-                if buf == b'':
-                    break
-                hash_func.update(buf)
-            f.close()
-            return hash_func.hexdigest() == digest
-        return False
+        if not os.path.isfile(pathname):
+            return False
+        alg, checksums = self.get_strongest_checksum()
+        size, digest = checksums[os.path.basename(pathname)]
+        return verify_file_checksum(pathname, alg, digest, size)
 
     def compare_dsc(self, other):
         """Check whether any files in these two dscs that have the same name
@@ -298,17 +288,6 @@ class SourcePackage(object):
         return os.path.join(mirror, 'pool', self.component, group,
                             self.source, filename)
 
-    def _lp_url(self, filename, source=False):
-        "Build an archive file URL on Launchpad"
-        if source:
-            suffix = '+sourcefiles/%s/%s/%s' % (
-                quote(self.source), quote(self.version.full_version),
-                quote(filename))
-        else:
-            suffix = '+files/%s' % quote(filename)
-        return 'https://launchpad.net/%s/+archive/primary/%s' % (
-            quote(self.distribution), suffix)
-
     def _source_urls(self, name):
         "Generator of sources for name"
         if self._dsc_source:
@@ -318,23 +297,20 @@ class SourcePackage(object):
         for mirror in self.masters:
             if mirror not in self.mirrors:
                 yield self._mirror_url(mirror, name)
-        yield self._lp_url(name, source=True)
+        if self.lp_spph.sourceFileUrl(name):
+            yield self.lp_spph.sourceFileUrl(name)
 
-    def _binary_urls(self, name, default_urls):
+    def _binary_urls(self, name, bpph):
         "Generator of URLs for name"
         for mirror in self.mirrors:
             yield self._mirror_url(mirror, name)
         for mirror in self.masters:
             if mirror not in self.mirrors:
                 yield self._mirror_url(mirror, name)
-        yield self._lp_url(name)
-        for url in default_urls:
-            yield url
-
-    def _binary_files_info(self, arch, name, ext):
-        for bpph in self.lp_spph.getBinaries(arch=arch, name=name, ext=ext):
-            urls = bpph.binaryFileUrls() + [bpph.getUrl()]
-            yield (bpph.getFileName(), urls, 0)
+        if bpph.binaryFileUrl(name):
+            yield bpph.binaryFileUrl(name)
+        if bpph.getUrl():
+            yield bpph.getUrl()
 
     def pull_dsc(self):
         "Retrieve dscfile and parse"
@@ -343,11 +319,17 @@ class SourcePackage(object):
             if parsed.scheme == '':
                 self._dsc_source = 'file://' + os.path.abspath(self._dsc_source)
                 parsed = urlparse(self._dsc_source)
-            url = self._dsc_source
+            self._download_dsc(self._dsc_source)
         else:
-            url = self._lp_url(self.dsc_name, source=True)
-        self._download_dsc(url)
-
+            for url in self._source_urls(self.dsc_name):
+                dlerr = None
+                try:
+                    self._download_dsc(url)
+                    break
+                except DownloadError as e:
+                    dlerr = e
+            else:
+                raise dlerr
         self._check_dsc()
 
     def _download_dsc(self, url):
@@ -420,54 +402,26 @@ class SourcePackage(object):
         with open(self.dsc_pathname, 'wb') as f:
             f.write(self.dsc.raw_text)
 
-    def _download_file_helper(self, f, pathname, size):
-        "Perform the dowload."
-        BLOCKSIZE = 16 * 1024
+    def _verify_file(self, pathname, dscverify=False, sha1sum=False, sha256sum=False, size=0):
+        if not os.path.exists(pathname):
+            return False
+        if size and size != os.path.getsize(pathname):
+            return False
+        if dscverify and not self.dsc.verify_file(pathname):
+            return False
+        if sha1sum and not verify_file_checksum(pathname, 'SHA1', sha1sum, size):
+            return False
+        if sha256sum and not verify_file_checksum(pathname, 'SHA256', sha256sum, size):
+            return False
+        return True
 
-        with open(pathname, 'wb') as out:
-            if not (Logger.isEnabledFor(logging.INFO) and
-                    sys.stderr.isatty() and
-                    size):
-                shutil.copyfileobj(f, out, BLOCKSIZE)
-                return
-
-            XTRALEN = len('[] 99%')
-            downloaded = 0
-            bar_width = 60
-            term_width = os.get_terminal_size(sys.stderr.fileno())[0]
-            if term_width < bar_width + XTRALEN + 1:
-                bar_width = term_width - XTRALEN - 1
-
-            try:
-                while True:
-                    block = f.read(BLOCKSIZE)
-                    if not block:
-                        break
-                    out.write(block)
-                    downloaded += len(block)
-                    pct = float(downloaded) / size
-                    bar = ('=' * int(pct * bar_width))[:-1] + '>'
-                    fmt = '[{bar:<%d}]{pct:>3}%%\r' % bar_width
-                    sys.stderr.write(fmt.format(bar=bar, pct=int(pct * 100)))
-                    sys.stderr.flush()
-            finally:
-                sys.stderr.write(' ' * (bar_width + XTRALEN) + '\r')
-                if downloaded < size:
-                    Logger.error('Partial download: %0.3f MiB of %0.3f MiB' %
-                                 (downloaded / 1024.0 / 1024,
-                                  size / 1024.0 / 1024))
-
-    def _download_file(self, url, filename, verify=True, size=0):
+    def _download_file(self, url, filename, size, dscverify=False, sha1sum=None, sha256sum=None):
         "Download url to filename in workdir."
         pathname = os.path.join(self.workdir, filename)
-        if verify:
-            if self.dsc.verify_file(pathname):
-                Logger.debug('Using existing %s', filename)
-                return True
-            size = [entry['size'] for entry in self.dsc['Files']
-                    if entry['name'] == filename]
-            assert len(size) == 1
-            size = int(size[0])
+
+        if self._verify_file(pathname, dscverify, sha1sum, sha256sum, size):
+            Logger.debug('Using existing file %s', filename)
+            return True
 
         if urlparse(url).scheme in ["", "file"]:
             frompath = os.path.abspath(urlparse(url).path)
@@ -477,31 +431,9 @@ class SourcePackage(object):
                 Logger.info("Copying %s from %s" % (filename, frompath))
                 shutil.copyfile(frompath, pathname)
         else:
-            try:
-                with closing(urlopen(url)) as f:
-                    Logger.debug("Using URL '%s'", f.geturl())
-                    if not size:
-                        try:
-                            size = int(f.info().get('Content-Length'))
-                        except (AttributeError, TypeError, ValueError):
-                            pass
+            download(url, pathname, size)
 
-                    Logger.info('Downloading %s from %s%s' %
-                                (filename, urlparse(url).hostname,
-                                 ' (%0.3f MiB)' % (size / 1024.0 / 1024)
-                                 if size else ''))
-
-                    self._download_file_helper(f, pathname, size)
-            except HTTPError as e:
-                # It's ok if the file isn't found; we try multiple places to download
-                if e.code == 404:
-                    return False
-                raise e
-
-        if verify and not self.dsc.verify_file(pathname):
-            Logger.error('Checksum for %s does not match.', filename)
-            return False
-        return True
+        return self._verify_file(pathname, dscverify, sha1sum, sha256sum, size)
 
     def pull(self):
         "Pull into workdir"
@@ -510,9 +442,13 @@ class SourcePackage(object):
             name = entry['name']
             for url in self._source_urls(name):
                 try:
-                    if self._download_file(url, name):
+                    if self._download_file(url, name, int(entry['size']), dscverify=True):
                         break
                 except HTTPError as e:
+                    # It's ok if the file isn't found; we try multiple places to download
+                    if e.code == 404:
+                        Logger.info("File not found at %s" % url)
+                        continue
                     Logger.info('HTTP Error %i: %s', e.code, str(e))
                 except URLError as e:
                     Logger.info('URL Error: %s', e.reason)
@@ -537,19 +473,25 @@ class SourcePackage(object):
         if arch == 'all':
             arch = None
 
-        for (fname, furls, fsize) in self._binary_files_info(arch, name, ext):
-            found = False
-            for url in self._binary_urls(fname, furls):
+        for bpph in self.lp_spph.getBinaries(arch=arch, name=name, ext=ext):
+            fname = bpph.getFileName()
+            fsha1 = bpph.binaryFileSha1(fname)
+            fsha256 = bpph.binaryFileSha256(fname)
+            fsize = bpph.binaryFileSize(fname)
+            for url in self._binary_urls(fname, bpph):
                 try:
-                    if self._download_file(url, fname, False, fsize):
-                        found = True
+                    if self._download_file(url, fname, fsize,
+                                           sha1sum=fsha1, sha256sum=fsha256):
+                        total += 1
                         break
                 except HTTPError as e:
+                    # It's ok if the file isn't found; we try multiple places to download
+                    if e.code == 404:
+                        Logger.info("File not found at %s" % url)
+                        continue
                     Logger.info('HTTP Error %i: %s', e.code, str(e))
                 except URLError as e:
                     Logger.info('URL Error: %s', e.reason)
-            if found:
-                total += 1
             else:
                 Logger.info("Could not download from any location: %s", fname)
         return total
@@ -609,11 +551,16 @@ class DebianSPPH(SourcePackagePublishingHistory):
     """
     resource_type = 'source_package_publishing_history'
 
+    def __init__(self, *args, **kwargs):
+        super(DebianSPPH, self).__init__(*args, **kwargs)
+        self._srcpkg = None
+
     def getBinaries(self, arch=None, name=None, ext=None):
-        Logger.info('Using Snapshot to find binary packages')
-        srcpkg = Snapshot.getSourcePackage(self.getPackageName(),
-                                           version=self.getVersion())
-        return srcpkg.getSPPH().getBinaries(arch=arch, name=name, ext=ext)
+        if not self._srcpkg:
+            Logger.info('Using Snapshot to find binary packages')
+            self._srcpkg = Snapshot.getSourcePackage(self.getPackageName(),
+                                                     version=self.getVersion())
+        return self._srcpkg.getSPPH().getBinaries(arch=arch, name=name, ext=ext)
 
 
 class DebianSourcePackage(SourcePackage):
@@ -665,17 +612,10 @@ class DebianSourcePackage(SourcePackage):
 
     def _source_urls(self, name):
         "Generator of sources for name"
-        wrapped_iterator = super(DebianSourcePackage, self)._source_urls(name)
-        while True:
-            try:
-                yield next(wrapped_iterator)
-            except StopIteration:
-                break
-        yield self.snapshot_files[name]
-
-    def _binary_files_info(self, arch, name, ext):
-        for f in self.snapshot_package.getBinaryFiles(arch=arch, name=name, ext=ext):
-            yield (f.name, [f.getUrl()], f.size)
+        for url in super(DebianSourcePackage, self)._source_urls(name):
+            yield url
+        if name in self.snapshot_files:
+            yield self.snapshot_files[name]
 
     def pull_dsc(self):
         "Retrieve dscfile and parse"
@@ -765,18 +705,6 @@ class PersonalPackageArchiveSourcePackage(UbuntuSourcePackage):
         self._team = None
         self._ppa = None
 
-    def _lp_url(self, filename, source=False):
-        "Build an archive file URL on Launchpad"
-        if source:
-            suffix = '+sourcefiles/%s/%s/%s' % (
-                quote(self.source), quote(self.version.full_version),
-                quote(filename))
-        else:
-            suffix = '+files/%s' % quote(filename)
-        return 'https://launchpad.net/~%s/+archive/%s/%s/%s' % (
-            quote(self._ppateam), quote(self.distribution),
-            quote(self._ppaname), suffix)
-
 
 class UbuntuCloudArchiveSourcePackage(PersonalPackageArchiveSourcePackage):
     "Download / unpack an Ubuntu Cloud Archive source package"
@@ -1097,7 +1025,7 @@ class SnapshotFile(object):
 
     @property
     def size(self):
-        return self._obj['size']
+        return int(self._obj['size'])
 
     @property
     def date(self):
@@ -1191,6 +1119,37 @@ class SnapshotSPPH(object):
     def getComponent(self):
         return self._pkg.component
 
+    def sourceFileUrls(self, include_meta=False):
+        if include_meta:
+            return [{'url': f.getUrl(),
+                     'filename': f.name,
+                     'sha1': f.getHash(),
+                     'sha256': None,
+                     'size': f.size}
+                    for f in self._pkg.getFiles()]
+        return [f.getUrl() for f in self._pkg.getFiles()]
+
+    def sourceFileUrl(self, filename):
+        for f in self.sourceFileUrls(include_meta=True):
+            if filename == f['filename']:
+                return f['url']
+        return None
+
+    def sourceFileSha1(self, url_or_filename):
+        for f in self.sourceFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return f['sha1']
+        return None
+
+    def sourceFileSha256(self, url_or_filename):
+        return None
+
+    def sourceFileSize(self, url_or_filename):
+        for f in self.sourceFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return int(f['size'])
+        return 0
+
     def getChangelog(self, since_version=None):
         '''
         Return the changelog, optionally since a particular version
@@ -1285,9 +1244,33 @@ class SnapshotBPPH(object):
     def getComponent(self):
         return self._file.component
 
-    def binaryFileUrls(self):
+    def binaryFileUrls(self, include_meta=False):
+        if include_meta:
+            return [{'url': self.getUrl(),
+                     'filename': self.getFileName(),
+                     'sha1': self._file.getHash(),
+                     'sha256': None,
+                     'size': self._file.size}]
         return [self.getUrl()]
 
+    def binaryFileUrl(self, filename):
+        if filename == self.getFileName():
+            return self.getUrl()
+        return None
+
+    def binaryFileSha1(self, url_or_filename):
+        if url_or_filename in [self.getUrl(), self.getFileName()]:
+            return self._file.getHash()
+        return None
+
+    def binaryFileSha256(self, url_or_filename):
+        return None
+
+    def binaryFileSize(self, url_or_filename):
+        if url_or_filename in [self.getUrl(), self.getFileName()]:
+            return int(self._file.size)
+        return 0
+
     def getBuild(self):
         return None
 
