@@ -26,13 +26,17 @@
 # httplib2.debuglevel = 1
 
 import collections
+import os
 import re
 
+from copy import copy
+
 from debian.changelog import Changelog
 from httplib2 import Http, HttpLib2Error
 from launchpadlib.launchpad import Launchpad as LP
 from launchpadlib.errors import HTTPError
 from lazr.restfulclient.resource import Entry
+from urllib.parse import urlparse
 
 from ubuntutools.version import Version
 from ubuntutools.lp import (service, api_version)
@@ -59,6 +63,7 @@ __all__ = [
     'DistroSeries',
     'DistroArchSeries',
     'Launchpad',
+    'PackageUpload',
     'PersonTeam',
     'SourcePackagePublishingHistory',
     ]
@@ -346,6 +351,78 @@ class DistroSeries(BaseWrapper):
                 raise ArchSeriesNotFoundException(message)
         return self._architectures[archtag]
 
+    def getPackageUploads(self, name=None, pocket=None, version=None,
+                          status='Unapproved'):
+        '''Returns a list of PackageUploads for this series.'''
+        params = {'exact_match': True}
+        if name:
+            params['name'] = name
+        if pocket:
+            params['pocket'] = pocket
+        if version:
+            params['version'] = version
+        if status:
+            params['status'] = status
+        return [PackageUpload(p) for p in self._lpobject.getPackageUploads(**params)]
+
+
+class PackageUpload(BaseWrapper):
+    '''
+    Wrapper class around a LP package_upload object.
+    '''
+    resource_type = 'package_upload'
+
+    def __init__(self, *args):
+        self._custom_urls = None
+        self._source_urls = None
+        self._binary_urls = None
+        self._binary_properties = None
+        self._binary_prop_dict = None
+
+    def getArchive(self):
+        return Archive(self._lpobject.archive_link)
+
+    def getSourceArchive(self):
+        if self._lpobject.copy_source_archive_link:
+            return Archive(self._lpobject.copy_source_archive_link)
+        return None
+
+    def getDistroSeries(self):
+        return DistroSeries(self._lpobject.distroseries_link)
+
+    def changesFileUrl(self):
+        return self._lpobject.changes_file_url
+
+    def customFileUrls(self):
+        if not self._custom_urls:
+            self._custom_urls = self._lpobject.customFileUrls()
+        return copy(self._custom_urls)
+
+    def sourceFileUrls(self):
+        if not self._source_urls:
+            self._source_urls = self._lpobject.sourceFileUrls()
+        return copy(self._source_urls)
+
+    def binaryFileUrls(self):
+        if not self._binary_urls:
+            self._binary_urls = self._lpobject.binaryFileUrls()
+        return copy(self._binary_urls)
+
+    def getBinaryProperties(self):
+        if not self._binary_properties:
+            self._binary_properties = self._lpobject.getBinaryProperties()
+        return copy(self._binary_properties)
+
+    def binaryFileProperties(self, filename_or_url):
+        if not self._binary_prop_dict:
+            urls = self.binaryFileUrls()
+            props = self.getBinaryProperties()
+            self._binary_prop_dict = dict(zip(urls, props))
+            for (k, v) in copy(self._binary_prop_dict).items():
+                filename = os.path.basename(urlparse(k).path)
+                self._binary_prop_dict[filename] = v
+        return self._binary_prop_dict.get(filename_or_url, {})
+
 
 class Archive(BaseWrapper):
     '''
@@ -700,6 +777,7 @@ class SourcePackagePublishingHistory(BaseWrapper):
         self._changelog = None
         self._binaries = {}
         self._distro_series = None
+        self._source_urls = None
         # Don't share _builds between different
         # SourcePackagePublishingHistory objects
         if '_builds' not in self.__dict__:
@@ -794,6 +872,84 @@ class SourcePackagePublishingHistory(BaseWrapper):
             new_entries.append(str(block))
         return ''.join(new_entries)
 
+    def sourceFileUrls(self, include_meta=False):
+        '''
+        Return the URL for this source publication's files.
+
+        The include_meta param changes the return value;
+        when it is False (the default), an array of url strings is
+        returned.  When include_meta is True, an array is returned
+        with dicts, containing the entries:
+          url: the url string
+          sha1: the SHA1 checksum of the source file (if provided)
+          sha256: the SHA256 checksum of the source file
+          size: the size of the source file
+        Also, this function adds a 'filename' field:
+          filename: the filename parsed from the url path
+        '''
+        if not self._source_urls:
+            urls = self._lpobject.sourceFileUrls(include_meta=True)
+            if not urls:
+                Logger.warning('SPPH %s_%s has no sourceFileUrls' %
+                               (self.getPackageName(), self.getVersion()))
+            for u in urls:
+                u['filename'] = os.path.basename(urlparse(u['url']).path)
+            self._source_urls = urls
+
+        if include_meta:
+            return list(self._source_urls)
+        return [f['url'] for f in self._source_urls]
+
+    def sourceFileUrl(self, filename):
+        '''
+        Returns the URL for the specified source filename.
+
+        If the filename is not found in the sourceFileUrls(), this returns None.
+        '''
+        for f in self.sourceFileUrls(include_meta=True):
+            if filename == f['filename']:
+                return f['url']
+        return None
+
+    def sourceFileSha1(self, url_or_filename):
+        '''
+        Returns the SHA1 checksum for the specified source file url.
+
+        If the url is not found in the sourceFileUrls(), this returns None.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.sourceFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return f['sha1']
+        return None
+
+    def sourceFileSha256(self, url_or_filename):
+        '''
+        Returns the SHA256 checksum for the specified source file url.
+
+        If the url is not found in the sourceFileUrls(), this returns None.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.sourceFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return f['sha256']
+        return None
+
+    def sourceFileSize(self, url_or_filename):
+        '''
+        Returns the size for the specified source file url.
+
+        If the url is not found in the sourceFileUrls(), this returns 0.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.sourceFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return int(f['size'])
+        return 0
+
     def getBinaries(self, arch=None, name=None, ext=None):
         '''
         Returns the resulting BinaryPackagePublishingHistorys.
@@ -824,7 +980,7 @@ class SourcePackagePublishingHistory(BaseWrapper):
             urls = self.binaryFileUrls()
             for url in urls:
                 # strip out the URL leading text.
-                filename = url.rsplit('/', 1)[1]
+                filename = os.path.basename(urlparse(url).path)
                 # strip the file suffix
                 (pkgname, _, e) = filename.rpartition('.')
                 # split into name, version, arch
@@ -936,6 +1092,7 @@ class BinaryPackagePublishingHistory(BaseWrapper):
     def __init__(self, *args):
         self._arch = None
         self._ext = None
+        self._binary_urls = None
 
     @property
     def arch(self):
@@ -968,20 +1125,89 @@ class BinaryPackagePublishingHistory(BaseWrapper):
         '''
         return self._lpobject.component_name
 
-    def binaryFileUrls(self):
+    def binaryFileUrls(self, include_meta=False):
         '''
         Return the URL for this binary publication's files.
         Only available in the devel API, not 1.0
+
+        The include_meta param changes the return value;
+        when it is False (the default), an array of url strings is
+        returned (but typically there is only a single url in the array).
+        When include_meta is True, an array (again, with typically only one
+        entry) is returned with dicts, containing the entries:
+          url: the url string
+          sha1: the SHA1 checksum of the binary file
+          sha256: the SHA256 checksum of the binary file
+          size: the size of the binary file
+        Also, this function adds a 'filename' field:
+          filename: the filename parsed from the url path
         '''
-        try:
-            urls = self._lpobject.binaryFileUrls()
+        if not self._binary_urls:
+            try:
+                urls = self._lpobject.binaryFileUrls(include_meta=True)
+            except AttributeError:
+                raise AttributeError("binaryFileUrls can only be found in lpapi "
+                                     "devel, not 1.0. Login using devel to have it.")
             if not urls:
                 Logger.warning('BPPH %s_%s has no binaryFileUrls' %
                                (self.getPackageName(), self.getVersion()))
-            return urls
-        except AttributeError:
-            raise AttributeError("binaryFileUrls can only be found in lpapi "
-                                 "devel, not 1.0. Login using devel to have it.")
+            for u in urls:
+                u['filename'] = os.path.basename(urlparse(u['url']).path)
+            self._binary_urls = urls
+
+        if include_meta:
+            return list(self._binary_urls)
+        return [f['url'] for f in self._binary_urls]
+
+    def binaryFileUrl(self, filename):
+        '''
+        Returns the URL for the specified binary filename.
+
+        If the filename is not found in the binaryFileUrls(), this returns None.
+        '''
+        for f in self.binaryFileUrls(include_meta=True):
+            if filename == f['filename']:
+                return f['url']
+        return None
+
+    def binaryFileSha1(self, url_or_filename):
+        '''
+        Returns the SHA1 checksum for the specified binary file url.
+
+        If the url is not found in the binaryFileUrls(), this returns None.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.binaryFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return f['sha1']
+        return None
+
+    def binaryFileSha256(self, url_or_filename):
+        '''
+        Returns the SHA256 checksum for the specified binary file url.
+
+        If the url is not found in the binaryFileUrls(), this returns None.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.binaryFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return f['sha256']
+        return None
+
+    def binaryFileSize(self, url_or_filename):
+        '''
+        Returns the size for the specified binary file url.
+
+        If the url is not found in the binaryFileUrls(), this returns 0.
+
+        The url may be specified as a filename.
+        '''
+        for f in self.binaryFileUrls(include_meta=True):
+            if url_or_filename in [f['url'], f['filename']]:
+                return int(f['size'])
+        return 0
 
     def getBuild(self):
         '''
