@@ -22,22 +22,31 @@
 #
 # ##################################################################
 
-# Modules.
+import distro_info
+import hashlib
 import locale
 import os
+import shutil
 import sys
-from subprocess import check_output, CalledProcessError
 
-import distro_info
+from subprocess import check_output, CalledProcessError
+from urllib.parse import urlparse
+from urllib.request import urlopen
 
 from ubuntutools.lp.udtexceptions import PocketDoesNotExistError
 
+import logging
+Logger = logging.getLogger(__name__)
+
+
 DEFAULT_POCKETS = ('Release', 'Security', 'Updates', 'Proposed')
 POCKETS = DEFAULT_POCKETS + ('Backports',)
 
 DEFAULT_STATUSES = ('Pending', 'Published')
 STATUSES = DEFAULT_STATUSES + ('Superseded', 'Deleted', 'Obsolete')
 
+UPLOAD_QUEUE_STATUSES = ('New', 'Unapproved', 'Accepted', 'Done', 'Rejected')
+
 _system_distribution_chain = []
 
 
@@ -57,7 +66,7 @@ def system_distribution_chain():
                                   encoding='utf-8').strip()
             _system_distribution_chain.append(vendor)
         except CalledProcessError:
-            print('Error: Could not determine what distribution you are running.')
+            Logger.error('Could not determine what distribution you are running.')
             return []
 
         while True:
@@ -97,8 +106,8 @@ def host_architecture():
         arch = None
 
     if not arch or 'not found' in arch:
-        print('Error: Not running on a Debian based system; '
-              'could not detect its architecture.')
+        Logger.error('Not running on a Debian based system; '
+                     'could not detect its architecture.')
         return None
 
     return arch
@@ -112,14 +121,14 @@ def readlist(filename, uniq=True):
     """
 
     if not os.path.isfile(filename):
-        print('File "%s" does not exist.' % filename)
+        Logger.error('File "%s" does not exist.' % filename)
         return False
 
     with open(filename) as f:
         content = f.read().replace('\n', ' ').replace(',', ' ')
 
     if not content.strip():
-        print('File "%s" is empty.' % filename)
+        Logger.error('File "%s" is empty.' % filename)
         return False
 
     items = [item for item in content.split() if item]
@@ -156,7 +165,7 @@ def split_release_pocket(release, default='Release'):
 def require_utf8():
     '''Can be called by programs that only function in UTF-8 locales'''
     if locale.getpreferredencoding() != 'UTF-8':
-        print("This program only functions in a UTF-8 locale. Aborting.", file=sys.stderr)
+        Logger.error("This program only functions in a UTF-8 locale. Aborting.")
         sys.exit(1)
 
 
@@ -188,3 +197,83 @@ def codename_to_distribution(codename):
 
         if info().valid(codename):
             return distro
+
+
+def verify_file_checksum(pathname, alg, checksum, size=0):
+    if not os.path.isfile(pathname):
+        Logger.error('File not found: %s', pathname)
+        return False
+    filename = os.path.basename(pathname)
+    if size and size != os.path.getsize(pathname):
+        Logger.error('File %s incorrect size, got %s expected %s',
+                     filename, os.path.getsize(pathname), size)
+        return False
+    h = hashlib.new(alg)
+    with open(pathname, 'rb') as f:
+        while True:
+            block = f.read(h.block_size)
+            if len(block) == 0:
+                break
+            h.update(block)
+    match = h.hexdigest() == checksum
+    if match:
+        Logger.debug('File %s checksum (%s) verified: %s',
+                     filename, alg, checksum)
+    else:
+        Logger.error('File %s checksum (%s) mismatch: got %s expected %s',
+                     filename, alg, h.hexdigest(), checksum)
+    return match
+
+
+def download(src, dst=None, size=0):
+    """ download/copy a file/url to local file """
+    filename = os.path.basename(urlparse(src).path)
+
+    if not dst:
+        dst = filename
+
+    with urlopen(src) as fsrc, open(dst, 'wb') as fdst:
+        url = fsrc.geturl()
+        Logger.debug(f"Using URL: {url}")
+
+        if not size:
+            try:
+                size = int(fsrc.info().get('Content-Length'))
+            except (AttributeError, TypeError, ValueError):
+                pass
+
+        hostname = urlparse(url).hostname
+        sizemb = ' (%0.3f MiB)' % (size / 1024.0 / 1024) if size else ''
+        Logger.info(f'Downloading {filename} from {hostname}{sizemb}')
+
+        if not all((Logger.isEnabledFor(logging.INFO),
+                    sys.stderr.isatty(), size)):
+            shutil.copyfileobj(fsrc, fdst)
+            return
+
+        blocksize = 4096
+        XTRALEN = len('[] 99%')
+        downloaded = 0
+        bar_width = 60
+        term_width = os.get_terminal_size(sys.stderr.fileno())[0]
+        if term_width < bar_width + XTRALEN + 1:
+            bar_width = term_width - XTRALEN - 1
+
+        try:
+            while True:
+                block = fsrc.read(blocksize)
+                if not block:
+                    break
+                fdst.write(block)
+                downloaded += len(block)
+                pct = float(downloaded) / size
+                bar = ('=' * int(pct * bar_width))[:-1] + '>'
+                fmt = '\r[{bar:<%d}]{pct:>3}%%\r' % bar_width
+                sys.stderr.write(fmt.format(bar=bar, pct=int(pct * 100)))
+                sys.stderr.flush()
+        finally:
+            sys.stderr.write('\r' + ' ' * (term_width - 1) + '\r')
+            if downloaded < size:
+                Logger.error('Partial download: %0.3f MiB of %0.3f MiB' %
+                             (downloaded / 1024.0 / 1024,
+                              size / 1024.0 / 1024))
